# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from sqlalchemy import create_engine

from datahub.emitter.mce_builder import dataset_urn_to_key
from datahub.ingestion.api.closeable import Closeable
from datahub.ingestion.source.snowflake.snowflake_config import SnowflakeConfig
from datahub.metadata.schema_classes import GlossaryNodeInfoClass, GlossaryTermInfoClass
from datahub.utilities.urns.urn import Urn
from datahub_actions.api.action_graph import AcrylDataHubGraph

logger: logging.Logger = logging.getLogger(__name__)


class SnowflakeTagHelper(Closeable):
    def __init__(self, config: SnowflakeConfig):
        self.config: SnowflakeConfig = config
        url = self.config.get_sql_alchemy_url()
        self.engine = create_engine(url, **self.config.get_options())

    @staticmethod
    def get_term_name_from_id(term_urn: str, graph: AcrylDataHubGraph) -> str:
        term_id = Urn.from_string(term_urn).get_entity_id_as_string()
        if term_id.count("-") == 4:
            # needs resolution
            term_info = graph.graph.get_aspect(term_urn, GlossaryTermInfoClass)
            assert term_info
            assert term_info.name
            term_name = term_info.name
            parent = term_info.parentNode
            while parent:
                parent_id = Urn.from_string(parent).get_entity_id_as_string()
                node_info = graph.graph.get_aspect(parent, GlossaryNodeInfoClass)
                assert node_info
                if parent_id.count("-") == 4:
                    parent_name = node_info.name
                    parent = node_info.parentNode
                else:
                    # terminate
                    parent_name = parent_id
                    parent = None
                term_name = f"{parent_name}.{term_name}"
        else:
            term_name = term_id

        return term_name

    @staticmethod
    def get_label_urn_to_tag(label_urn: str, graph: AcrylDataHubGraph) -> str:
        label_urn_parsed = Urn.from_string(label_urn)
        if label_urn_parsed.get_type() == "tag":
            return label_urn_parsed.get_entity_id_as_string()
        elif label_urn_parsed.get_type() == "glossaryTerm":
            # if this looks like a guid, we want to resolve to human friendly names
            term_name = SnowflakeTagHelper.get_term_name_from_id(label_urn, graph)
            if term_name is not None:
                # terms use `.` for separation, replace with _
                return term_name.replace(".", "_").replace(" ", "_")
            else:
                raise ValueError(f"Invalid tag or term urn {label_urn}")
        else:
            raise Exception(
                f"Unexpected label type: neither tag or term {label_urn_parsed.get_type()}"
            )

    def apply_tag_or_term(
        self, dataset_urn: str, tag_or_term_urn: str, graph: AcrylDataHubGraph
    ) -> None:
        dataset_key = dataset_urn_to_key(dataset_urn)
        assert dataset_key is not None
        if dataset_key.platform != "snowflake":
            return
        tag = self.get_label_urn_to_tag(tag_or_term_urn, graph)
        assert tag is not None
        name_tokens = dataset_key.name.split(".")
        assert len(name_tokens) == 3
        self.run_query(
            name_tokens[0],
            name_tokens[1],
            f"CREATE TAG IF NOT EXISTS {tag} COMMENT = 'Replicated Tag {tag_or_term_urn} from DataHub';",
        )
        self.run_query(
            name_tokens[0],
            name_tokens[1],
            f'ALTER TABLE {name_tokens[2]} SET TAG {tag}="{tag_or_term_urn}";',
        )

    def remove_tag_or_term(
        self, dataset_urn: str, tag_urn: str, graph: AcrylDataHubGraph
    ) -> None:
        dataset_key = dataset_urn_to_key(dataset_urn)
        assert dataset_key is not None
        if dataset_key.platform != "snowflake":
            return
        tag = self.get_label_urn_to_tag(tag_urn, graph)
        assert tag is not None
        name_tokens = dataset_key.name.split(".")
        assert len(name_tokens) == 3
        self.run_query(
            name_tokens[0],
            name_tokens[1],
            f"ALTER TABLE {name_tokens[2]} UNSET TAG {tag};",
        )

    def run_query(self, database: str, schema: str, query: str) -> None:
        try:
            self.engine.execute(f"USE {database}.{schema};")
            self.engine.execute(query)
            logger.info(f"Successfully executed query {query}")
        except Exception as e:
            logger.warning(
                f"Failed to execute snowflake query: {query}. Exception: ", e
            )

    def close(self) -> None:
        return



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

from datahub.configuration import ConfigModel
from datahub.configuration.common import ConfigEnum
from datahub.ingestion.graph.client import DatahubClientConfig


class FailureMode(ConfigEnum):
    # Log the failed event to the failed events log. Then throw an pipeline exception to stop the pipeline.
    THROW = "THROW"
    # Log the failed event to the failed events log. Then continue processing the event stream.
    CONTINUE = "CONTINUE"


class SourceConfig(ConfigModel):
    type: str
    config: Optional[Dict[str, Any]] = None


class TransformConfig(ConfigModel):
    type: str
    config: Optional[Dict[str, Any]] = None


class FilterConfig(ConfigModel):
    event_type: Union[str, List[str]]
    event: Optional[Dict[str, Any]] = None


class ActionConfig(ConfigModel):
    type: str
    config: Optional[dict]


class PipelineOptions(BaseModel):
    retry_count: Optional[int] = None
    failure_mode: Optional[FailureMode] = None
    failed_events_dir: Optional[str] = (
        None  # The path where failed events should be logged.
    )


class PipelineConfig(ConfigModel):
    """
    Configuration required to create a new Actions Pipeline.

    This exactly matches the structure of the YAML file used
    to configure a Pipeline.
    """

    name: str
    enabled: bool = True
    source: SourceConfig
    filter: Optional[FilterConfig] = None
    transform: Optional[List[TransformConfig]] = None
    action: ActionConfig
    datahub: Optional[DatahubClientConfig] = None
    options: Optional[PipelineOptions] = None



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import List, Optional

from datahub_actions.action.action import Action
from datahub_actions.event.event_envelope import EventEnvelope
from datahub_actions.pipeline.pipeline_config import FailureMode, PipelineConfig
from datahub_actions.pipeline.pipeline_stats import PipelineStats
from datahub_actions.pipeline.pipeline_util import (
    create_action,
    create_action_context,
    create_event_source,
    create_filter_transformer,
    create_transformer,
    normalize_directory_name,
)
from datahub_actions.source.event_source import EventSource
from datahub_actions.transform.transformer import Transformer

logger = logging.getLogger(__name__)


# Defaults for the location where failed events will be written.
DEFAULT_RETRY_COUNT = 0  # Do not retry unless instructed.
DEFAULT_FAILED_EVENTS_DIR = "/tmp/logs/datahub/actions"
DEFAULT_FAILED_EVENTS_FILE_NAME = "failed_events.log"  # Not currently configurable.
DEFAULT_FAILURE_MODE = FailureMode.CONTINUE


class PipelineException(Exception):
    """
    An exception thrown when a Pipeline encounters and unrecoverable situation.
    Mainly a placeholder for now.
    """

    pass


class Pipeline:
    """
    A Pipeline is responsible for coordinating execution of a single DataHub Action.

    This responsibility includes:

        - sourcing events from an Event Source
        - executing a configurable chain of Transformers
        - invoking an Action with the final Event
        - acknowledging the processing of an Event with the Event Source

    Additionally, a Pipeline supports the following notable capabilities:

        - Configurable retries of event processing in cases of component failure
        - Configurable dead letter queue
        - Capturing basic statistics about each Pipeline component
        - At-will start and stop of an individual pipeline

    """

    name: str
    source: EventSource
    transforms: List[Transformer] = []
    action: Action

    # Whether the Pipeline has been requested to shut down
    _shutdown: bool = False

    # Pipeline statistics
    _stats: PipelineStats = PipelineStats()

    # Options
    _retry_count: int = DEFAULT_RETRY_COUNT  # Number of times a single event should be retried in case of processing error.
    _failure_mode: FailureMode = DEFAULT_FAILURE_MODE
    _failed_events_dir: str = DEFAULT_FAILED_EVENTS_DIR  # The top-level path where failed events will be logged.

    def __init__(
        self,
        name: str,
        source: EventSource,
        transforms: List[Transformer],
        action: Action,
        retry_count: Optional[int],
        failure_mode: Optional[FailureMode],
        failed_events_dir: Optional[str],
    ) -> None:
        self.name = name
        self.source = source
        self.transforms = transforms
        self.action = action

        if retry_count is not None:
            self._retry_count = retry_count
        if failure_mode is not None:
            self._failure_mode = failure_mode
        if failed_events_dir is not None:
            self._failed_events_dir = failed_events_dir
        self._init_failed_events_dir()

    @classmethod
    def create(cls, config_dict: dict) -> "Pipeline":
        # Bind config
        config = PipelineConfig.model_validate(config_dict)

        if not config.enabled:
            raise Exception(
                "Pipeline is disabled, but create method was called unexpectedly."
            )

        # Create Context
        ctx = create_action_context(config.name, config.datahub)

        # Create Event Source
        event_source = create_event_source(config.source, ctx)

        # Create Transforms
        transforms = []
        if config.filter is not None:
            transforms.append(create_filter_transformer(config.filter, ctx))

        if config.transform is not None:
            for transform_config in config.transform:
                transforms.append(create_transformer(transform_config, ctx))

        # Create Action
        action = create_action(config.action, ctx)

        # Finally, create Pipeline.
        return cls(
            config.name,
            event_source,
            transforms,
            action,
            config.options.retry_count if config.options else None,
            config.options.failure_mode if config.options else None,
            config.options.failed_events_dir if config.options else None,
        )

    async def start(self) -> None:
        """
        Start the action pipeline asynchronously. This method is non-blocking.
        """
        self.run()

    def run(self) -> None:
        """
        Run the action pipeline synchronously. This method is blocking.
        Raises an instance of PipelineException if an unrecoverable pipeline failure occurs.
        """
        self._stats.mark_start()

        # First, source the events.
        enveloped_events = self.source.events()
        for enveloped_event in enveloped_events:
            # Then, process the event.
            retval = self._process_event(enveloped_event)

            # For legacy users w/o selective ack support, convert
            # None to True, i.e. always commit.
            if retval is None:
                retval = True

            # Finally, ack the event.
            self._ack_event(enveloped_event, retval)

    def stop(self) -> None:
        """
        Stops a running action pipeline.
        """
        logger.debug(f"Preparing to stop Actions Pipeline with name {self.name}")
        self._shutdown = True
        self._failed_events_fd.close()
        self.source.close()
        self.action.close()

    def stats(self) -> PipelineStats:
        """
        Returns basic statistics about the Pipeline run.
        """
        return self._stats

    def _process_event(self, enveloped_event: EventEnvelope) -> Optional[bool]:
        # Attempt to process the incoming event, with retry.
        curr_attempt = 1
        max_attempts = self._retry_count + 1
        retval = None
        while curr_attempt <= max_attempts:
            try:
                # First, transform the event.
                transformed_event = self._execute_transformers(enveloped_event)

                # Then, invoke the action if the event is non-null.
                if transformed_event is not None:
                    retval = self._execute_action(transformed_event)

                # Short circuit - processing has succeeded.
                return retval
            except Exception:
                logger.exception(
                    f"Caught exception while attempting to process event. Attempt {curr_attempt}/{max_attempts} event type: {enveloped_event.event_type}, pipeline name: {self.name}"
                )
                curr_attempt = curr_attempt + 1

        logger.error(
            f"Failed to process event after {self._retry_count} retries. event type: {enveloped_event.event_type}, pipeline name: {self.name}. Handling failure..."
        )

        # Increment failed event count.
        self._stats.increment_failed_event_count()

        # Finally, handle the failure
        self._handle_failure(enveloped_event)

        return retval

    def _execute_transformers(
        self, enveloped_event: EventEnvelope
    ) -> Optional[EventEnvelope]:
        curr_event = enveloped_event
        # Iterate through all transformers, sequentially apply them to the result of the previous.
        for transformer in self.transforms:
            # Increment stats
            self._stats.increment_transformer_processed_count(transformer)

            # Transform the event
            transformed_event = self._execute_transformer(curr_event, transformer)

            # Process result
            if transformed_event is None:
                # If the transformer has filtered the event, short circuit.
                self._stats.increment_transformer_filtered_count(transformer)
                return None
            # Otherwise, set the result to the transformed event.
            curr_event = transformed_event

        # Return the final transformed event.
        return curr_event

    def _execute_transformer(
        self, enveloped_event: EventEnvelope, transformer: Transformer
    ) -> Optional[EventEnvelope]:
        try:
            return transformer.transform(enveloped_event)
        except Exception as e:
            self._stats.increment_transformer_exception_count(transformer)
            raise PipelineException(
                f"Caught exception while executing Transformer with name {type(transformer).__name__}"
            ) from e

    def _execute_action(self, enveloped_event: EventEnvelope) -> Optional[bool]:
        try:
            retval = self.action.act(enveloped_event)
            self._stats.increment_action_success_count()
            return retval
        except Exception as e:
            self._stats.increment_action_exception_count()
            raise PipelineException(
                f"Caught exception while executing Action with type {type(self.action).__name__}"
            ) from e

    def _ack_event(self, enveloped_event: EventEnvelope, processed: bool) -> None:
        try:
            self.source.ack(enveloped_event, processed)
            self._stats.increment_success_count()
        except Exception:
            self._stats.increment_failed_ack_count()
            logger.exception(
                f"Caught exception while attempting to ack successfully processed event. event type: {enveloped_event.event_type}, pipeline name: {self.name}",
            )
            logger.debug(f"Failed to ack event: {enveloped_event}")

    def _handle_failure(self, enveloped_event: EventEnvelope) -> None:
        # First, always save the failed event to a file. Useful for investigation.
        self._append_failed_event_to_file(enveloped_event)
        if self._failure_mode == FailureMode.THROW:
            raise PipelineException("Failed to process event after maximum retries.")
        elif self._failure_mode == FailureMode.CONTINUE:
            # Simply return, nothing left to do.
            pass

    def _append_failed_event_to_file(self, enveloped_event: EventEnvelope) -> None:
        # First, convert the event to JSON.
        try:
            json = enveloped_event.as_json()
            # Then append to failed events file.
            self._failed_events_fd.write(json + "\n")
            self._failed_events_fd.flush()
        except Exception as e:
            # This is a serious issue, as if we do not handle it can mean losing an event altogether.
            # Raise an exception to ensure this issue is reported to the operator.
            raise PipelineException(
                f"Failed to log failed event to file! {enveloped_event}"
            ) from e

    def _init_failed_events_dir(self) -> None:
        # create a directory for failed events from this actions pipeine.
        failed_events_dir = os.path.join(
            self._failed_events_dir, normalize_directory_name(self.name)
        )
        try:
            os.makedirs(failed_events_dir, exist_ok=True)

            failed_events_file_name = os.path.join(
                failed_events_dir, DEFAULT_FAILED_EVENTS_FILE_NAME
            )
            self._failed_events_fd = open(failed_events_file_name, "a")
        except Exception as e:
            logger.debug(e)
            raise PipelineException(
                f"Caught exception while attempting to create failed events log file at path {failed_events_dir}. Please check your file system permissions."
            ) from e



import json
import logging
import time
from dataclasses import dataclass
from typing import Iterable, List, Optional

from datahub.configuration import ConfigModel
from datahub.emitter.serialization_helper import post_json_transform

# DataHub imports.
from datahub.metadata.schema_classes import GenericPayloadClass
from datahub_actions.event.event_envelope import EventEnvelope
from datahub_actions.event.event_registry import (
    ENTITY_CHANGE_EVENT_V1_TYPE,
    EntityChangeEvent,
)

# May or may not need these.
from datahub_actions.pipeline.pipeline_context import PipelineContext
from datahub_actions.plugin.source.acryl.constants import (
    ENTITY_CHANGE_EVENT_NAME,
    PLATFORM_EVENT_TOPIC_NAME,
)
from datahub_actions.plugin.source.acryl.datahub_cloud_events_ack_manager import (
    AckManager,
)
from datahub_actions.plugin.source.acryl.datahub_cloud_events_consumer import (
    DataHubEventsConsumer,
    ExternalEvent,
)
from datahub_actions.source.event_source import EventSource

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


# Converts a DataHub Events Message to an EntityChangeEvent.
def build_entity_change_event(payload: GenericPayloadClass) -> EntityChangeEvent:
    try:
        return EntityChangeEvent.from_json(payload.get("value"))
    except Exception as e:
        raise ValueError("Failed to parse into EntityChangeEvent") from e


class DataHubEventsSourceConfig(ConfigModel):
    topic: str = PLATFORM_EVENT_TOPIC_NAME
    consumer_id: Optional[str] = None  # Used to store offset for the consumer.
    lookback_days: Optional[int] = None
    reset_offsets: Optional[bool] = False

    # Time and Exit Conditions.
    kill_after_idle_timeout: bool = False
    idle_timeout_duration_seconds: int = 30
    event_processing_time_max_duration_seconds: int = 60


# This is the custom DataHub-based Event Source.
@dataclass
class DataHubEventSource(EventSource):
    running = False
    source_config: DataHubEventsSourceConfig
    ctx: PipelineContext

    @staticmethod
    def _get_pipeline_urn(pipeline_name: str) -> str:
        if pipeline_name.startswith("urn:li:dataHubAction:"):
            return pipeline_name
        else:
            return f"urn:li:dataHubAction:{pipeline_name}"

    def __init__(self, config: DataHubEventsSourceConfig, ctx: PipelineContext):
        self.ctx = ctx
        self.source_config = config
        self.consumer_id = DataHubEventSource._get_pipeline_urn(self.ctx.pipeline_name)

        # Ensure a Graph Instance was provided.
        assert self.ctx.graph is not None

        self.datahub_events_consumer: DataHubEventsConsumer = DataHubEventsConsumer(
            # TODO: This PipelineContext provides an Acryl Graph Instance
            graph=self.ctx.graph.graph,
            consumer_id=self.consumer_id,
            lookback_days=self.source_config.lookback_days,
            reset_offsets=self.source_config.reset_offsets,
        )
        self.ack_manager = AckManager()
        self.safe_to_ack_offset: Optional[str] = None

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "EventSource":
        config = DataHubEventsSourceConfig.model_validate(config_dict)
        return cls(config, ctx)

    def events(self) -> Iterable[EventEnvelope]:
        logger.info("Starting DataHub Cloud events source...")
        logger.info(f"Subscribing to the following topic: {self.source_config.topic}")
        self.running = True
        yield from self._poll_and_process_events()

    def _poll_and_process_events(self) -> Iterable[EventEnvelope]:
        """Poll and process events in the main loop."""
        last_idle_response_timestamp = 0
        while self.running:
            try:
                sleeps_to_go = (
                    self.source_config.event_processing_time_max_duration_seconds
                )

                while self.ack_manager.outstanding_acks():
                    time.sleep(1)
                    sleeps_to_go -= 1
                    logger.debug(f"Sleeps to go: {sleeps_to_go}")

                    if sleeps_to_go == 0:
                        self.running = False
                        raise Exception(
                            f"Failed to process all events successfully after specified time {self.source_config.event_processing_time_max_duration_seconds}! If more time is required, please increase the timeout using this config. {self.ack_manager.acks.values()}",
                        )
                logger.debug(
                    f"Successfully processed events up to offset id {self.safe_to_ack_offset}"
                )
                self.safe_to_ack_offset = self.datahub_events_consumer.offset_id
                logger.debug(f"Safe to ack offset: {self.safe_to_ack_offset}")

                events_response = self.datahub_events_consumer.poll_events(
                    topic=self.source_config.topic, poll_timeout_seconds=2
                )

                # Handle Idle Timeout
                num_events = len(events_response.events)

                if num_events == 0:
                    if last_idle_response_timestamp == 0:
                        last_idle_response_timestamp = (
                            self._get_current_timestamp_seconds()
                        )
                    if self._should_idle_timeout(
                        num_events, last_idle_response_timestamp
                    ):
                        logger.info("Exiting main loop due to idle timeout")
                        return
                else:
                    self.ack_manager.new_batch()
                    last_idle_response_timestamp = 0  # Reset the idle timeout

                event_envelopes: List[EventEnvelope] = []
                for msg in events_response.events:
                    for event_envelope in self.handle_pe(msg):
                        event_envelope.meta = self.ack_manager.get_meta(event_envelope)
                        event_envelopes.append(event_envelope)

                yield from event_envelopes

            except Exception as e:
                logger.exception(f"DataHub Events consumer error: {e}")
                self.running = False

        logger.info("DataHub Events consumer exiting main loop")

    @staticmethod
    def handle_pe(msg: ExternalEvent) -> Iterable[EventEnvelope]:
        value: dict = json.loads(msg.value)
        payload: GenericPayloadClass = GenericPayloadClass.from_obj(
            post_json_transform(value["payload"])
        )
        if ENTITY_CHANGE_EVENT_NAME == value["name"]:
            event = build_entity_change_event(payload)
            yield EventEnvelope(ENTITY_CHANGE_EVENT_V1_TYPE, event, {})

    def close(self) -> None:
        if self.datahub_events_consumer:
            self.running = False
            if self.safe_to_ack_offset:
                self.datahub_events_consumer.commit_offsets(
                    offset_id=self.safe_to_ack_offset
                )
            self.datahub_events_consumer.close()

    def ack(self, event: EventEnvelope, processed: bool = True) -> None:
        self.ack_manager.ack(event.meta, processed=processed)
        logger.debug(f"Actions acked event {event} as processed {processed}")

    def _should_idle_timeout(
        self, num_events: int, last_idle_response_timestamp: int
    ) -> bool:
        """Handle idle timeout logic and decide if the loop should exit."""
        if num_events > 0:
            return False  # Continue processing

        current_timestamp_seconds = self._get_current_timestamp_seconds()

        if (
            self.source_config.kill_after_idle_timeout
            and current_timestamp_seconds - last_idle_response_timestamp
            > self.source_config.idle_timeout_duration_seconds
        ):
            logger.info(
                f"Shutting down due to idle timeout of {self.source_config.idle_timeout_duration_seconds} seconds"
            )
            self.running = False
            return True  # Signal that we should exit
        return False  # Continue processing

    def _get_current_timestamp_seconds(self) -> int:
        return int(time.time())



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datahub.ingestion.api.registry import PluginRegistry
from datahub_actions.plugin.source.acryl.datahub_cloud_event_source import (
    DataHubEventSource,
)
from datahub_actions.plugin.source.kafka.kafka_event_source import KafkaEventSource
from datahub_actions.source.event_source import EventSource

event_source_registry = PluginRegistry[EventSource]()
event_source_registry.register_from_entrypoint("datahub_actions.source.plugins")
event_source_registry.register("kafka", KafkaEventSource)
event_source_registry.register("datahub-cloud", DataHubEventSource)



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import json
from datetime import datetime, timezone
from typing import Dict, Optional

import pydantic
from pydantic import BaseModel

from datahub.ingestion.api.report import Report, SupportsAsObj
from datahub.utilities.str_enum import StrEnum
from datahub_actions.action.action import Action
from datahub_actions.event.event_envelope import EventEnvelope
from datahub_actions.event.event_registry import (
    ENTITY_CHANGE_EVENT_V1_TYPE,
    METADATA_CHANGE_LOG_EVENT_V1_TYPE,
    EntityChangeEvent,
    MetadataChangeLogEvent,
)
from datahub_actions.pipeline.pipeline_context import PipelineContext


class EventProcessingStats(BaseModel):
    """
    A class to represent the event-oriented processing stats for a pipeline.
    Note: Might be merged into ActionStats in the future.
    """

    last_seen_event_time: Optional[str] = pydantic.Field(
        None, description="The event time of the last event we processed"
    )
    last_event_processed_time: Optional[str] = pydantic.Field(
        None, description="The time at which we processed the last event"
    )
    last_seen_event_time_success: Optional[str] = pydantic.Field(
        None, description="The event time of the last event we processed successfully"
    )
    last_event_processed_time_success: Optional[str] = pydantic.Field(
        None, description="The time at which we processed the last event successfully"
    )
    last_seen_event_time_failure: Optional[str] = pydantic.Field(
        None, description="The event time of the last event we processed unsuccessfully"
    )
    last_event_processed_time_failure: Optional[str] = pydantic.Field(
        None, description="The time at which we processed the last event unsuccessfully"
    )

    @classmethod
    def _get_event_time(cls, event: EventEnvelope) -> Optional[str]:
        """
        Get the event time from the event.
        """
        if event.event_type == ENTITY_CHANGE_EVENT_V1_TYPE:
            if isinstance(event.event, EntityChangeEvent):
                return (
                    datetime.fromtimestamp(
                        event.event.auditStamp.time / 1000.0, tz=timezone.utc
                    ).isoformat()
                    if event.event.auditStamp
                    else None
                )
        elif event.event_type == METADATA_CHANGE_LOG_EVENT_V1_TYPE:
            if isinstance(event.event, MetadataChangeLogEvent):
                return (
                    datetime.fromtimestamp(
                        event.event.auditHeader.time / 1000.0, tz=timezone.utc
                    ).isoformat()
                    if event.event.auditHeader
                    else None
                )
        return None

    def start(self, event: EventEnvelope) -> None:
        """
        Update the stats based on the event.
        """
        self.last_event_processed_time = datetime.now(tz=timezone.utc).isoformat()
        self.last_seen_event_time = self._get_event_time(event)

    def end(self, event: EventEnvelope, success: bool) -> None:
        """
        Update the stats based on the event.
        """

        if success:
            self.last_seen_event_time_success = (
                self._get_event_time(event) or self.last_seen_event_time_success
            )
            self.last_event_processed_time_success = datetime.now(
                timezone.utc
            ).isoformat()
        else:
            self.last_seen_event_time_failure = (
                self._get_event_time(event) or self.last_seen_event_time_failure
            )
            self.last_event_processed_time_failure = datetime.now(
                timezone.utc
            ).isoformat()

    def __str__(self) -> str:
        return json.dumps(self.model_dump(), indent=2)


class StageStatus(StrEnum):
    SUCCESS = "success"
    FAILURE = "failure"
    RUNNING = "running"
    STOPPED = "stopped"


class ActionStageReport(BaseModel):
    # All stats here are only for the current run of the current stage.

    # Attributes that should be aggregated across runs should be prefixed with "total_".
    # Only ints can be aggregated.

    start_time: int = 0

    end_time: int = 0

    total_assets_to_process: int = -1  # -1 if unknown

    total_assets_processed: int = 0

    total_actions_executed: int = 0

    total_assets_impacted: int = 0

    event_processing_stats: Optional[EventProcessingStats] = None

    status: Optional[StageStatus] = None

    def start(self) -> None:
        self.start_time = int(datetime.now().timestamp() * 1000)
        self.status = StageStatus.RUNNING

    def end(self, success: bool) -> None:
        self.end_time = int(datetime.now().timestamp() * 1000)
        self.status = StageStatus.SUCCESS if success else StageStatus.FAILURE

    def increment_assets_processed(self, asset: str) -> None:
        # TODO: If we want to track unique assets, use a counting set.
        # For now, just increment
        self.total_assets_processed += 1

    def increment_assets_impacted(self, asset: str) -> None:
        # TODO: If we want to track unique assets, use a counting set.
        # For now, just increment
        self.total_assets_impacted += 1

    def as_obj(self) -> dict:
        return Report.to_pure_python_obj(self)

    def aggregatable_stats(self) -> Dict[str, int]:
        all_items = self.model_dump()

        stats = {k: v for k, v in all_items.items() if k.startswith("total_")}

        # If total_assets_to_process is unknown, don't include it.
        if self.total_assets_to_process == -1:
            stats.pop("total_assets_to_process")

        # Add a few additional special cases of aggregatable stats.
        if self.event_processing_stats:
            for key, value in self.event_processing_stats.model_dump().items():
                if value is not None:
                    stats[f"event_processing_stats.{key}"] = str(value)

        return stats


class ReportingAction(Action, abc.ABC):
    def __init__(self, ctx: PipelineContext):
        super().__init__()
        self.ctx = ctx

        self.action_urn: str
        if "urn:li:dataHubAction:" in ctx.pipeline_name:
            # The pipeline name might get a prefix before the urn:li:... part.
            # We need to remove that prefix to get the urn:li:dataHubAction part.
            action_urn_part = ctx.pipeline_name.split("urn:li:dataHubAction:")[1]
            self.action_urn = f"urn:li:dataHubAction:{action_urn_part}"
        else:
            self.action_urn = f"urn:li:dataHubAction:{ctx.pipeline_name}"

    @abc.abstractmethod
    def get_report(self) -> ActionStageReport:
        pass


assert isinstance(ActionStageReport(), SupportsAsObj)
